﻿using Ceras;
using Ceras.Helpers;
using ChatObjects;
using Homework4Server;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ChatServer
{
	public class ClientHandler
    {
		TcpClient tcpClient;
		public NetworkStream netStream { get; }
		CerasSerializer sendCeras;
		CerasSerializer receiveCeras;
		bool stopListen = false;
		public string _clientName { get; set;  }
		public ReaderWriterLock readWriteLock;
		public List<ClientHandler> clientList { get; set; }

		public List<ObjectHandler> objectHandlers { get; set; }
		public ClientHandler(TcpClient tcpClient, ReaderWriterLock rwl, List<ObjectHandler> objectHandlers)
		{
			readWriteLock = rwl;
			this.clientList = clientList;
			this.objectHandlers = objectHandlers;
			this.tcpClient = tcpClient;
			this.netStream = tcpClient.GetStream();

			// We want to keep "learned" types
			// That means when the other side sends us a type (using the full name) we never want to transmit that again,
			// the type should (from then on) be known as a some ID.
			var config1 = new SerializerConfig();
			config1.Advanced.PersistTypeCache = false;

			var config2 = new SerializerConfig();
			config2.Advanced.PersistTypeCache = false;

			sendCeras = new CerasSerializer(config1);
			receiveCeras = new CerasSerializer(config2);

			StartReceivingMessages();
		}

		void StartReceivingMessages()
		{
			Task.Run(async () =>
			{
				try
				{
					// Keep receiving packets from the client and respond to them
					// Eventually when the client disconnects we'll just get an exception and end the thread...
					while (!stopListen)
					{
						var obj = await receiveCeras.ReadFromStream(netStream);
						HandleMessage(obj);
					}
				}
				catch (Exception e)
				{
					readWriteLock.AcquireWriterLock(-1);
					try
					{
						clientList.Remove(this);
					} finally
					{
						readWriteLock.ReleaseWriterLock();
					}
					//Log($"Error while handling client '{_tcpClient.Client.RemoteEndPoint}': {e}");
				}
			});
		}

		void HandleMessage(object obj)
		{
			for(int i = 0; i < objectHandlers.Count; i++)
			{
				if (objectHandlers[i].handleObject(obj,this))
				{
					break;
				}
			}
		}

		public void Log(string text) => Console.WriteLine("[Server] " + text);

		public void Send(Stream stream, object obj) {                    
			try
			{
				sendCeras.WriteToStream(stream, obj);
			}
			catch (Exception ex)
			{
				Log("Send message failed to client ["+_clientName+"]");
			}
		}
	
	}

	public static class Server
	{
		public static ReaderWriterLock readWriteLock { get; set; }
		static List<ClientHandler> clientList;
		static List<ObjectHandler> objectHandlers;
		static public void EndProgram()
		{
			ServerShutdown serverShutdown = new ServerShutdown();
			readWriteLock.AcquireReaderLock(-1);
			try
			{
				clientList.ForEach(client => client.Send(client.netStream, serverShutdown));
			} finally
			{
				readWriteLock.ReleaseReaderLock();
			}
		}

		public static void Start()
		{
			clientList = new List<ClientHandler>();
			readWriteLock = new ReaderWriterLock();
			objectHandlers = new List<ObjectHandler>();
			objectHandlers.Add(new StringHandler());
			objectHandlers.Add(new LoginHandler());
			objectHandlers.Add(new LogoutHandler());
			objectHandlers.Add(new GetUserListHandler());

			new Thread(AcceptClients).Start();
		}

		static void AcceptClients()
		{
			var listener = new TcpListener(IPAddress.Any, 43210);
			listener.Start();			
			while (true)
			{
				var tcpClient = listener.AcceptTcpClient();
				var clientHandler = new ClientHandler(tcpClient, readWriteLock, objectHandlers);
				readWriteLock.AcquireWriterLock(-1);
				try
				{
					clientList.Add(clientHandler);
				} finally
				{
					readWriteLock.ReleaseWriterLock();
				}
				
				clientHandler.clientList = clientList;
			}
		}

	}
}
